#include "gfx.h"

#define __PROG_TYPES_COMPAT__
#include <avr/pgmspace.h>
#include <avr/io.h>
#include <stdlib.h>
#include <util/delay.h>

#define NR_OBST	5
#define HEIGHT	128
#define WIDTH	128
#define LEFT	0
#define UP		1
#define RIGHT	2
#define DOWN	3
#define NONE	4

typedef struct{
	int x, y;
} snake_position;

typedef struct{
	int x1, y1, x2, y2;
} obstacle_coord;

typedef struct{
	int x, y;
} power_up;

snake_position snake[WIDTH];
obstacle_coord obstacles[NR_OBST];
int snake_length = 10;
int snake_direction = NONE;
int level = 2000;
power_up buff;
power_up normal_buff;

void buttons_init(){
	/* RIGHT */
	DDRD &= ~(1 << PD0);
	PORTD |= (1 << PD0);
	
	/* DOWN */
	DDRD &= ~(1 << PD1);
	PORTD |= (1 << PD1);
	
	/* LEFT */
	DDRD &= ~(1 << PD4);
	PORTD |= (1 << PD4);
	
	/* UP */
	DDRD &= ~(1 << PD5);
	PORTD |= (1 << PD5);
}

void new_power_up(){
	buff.x = rand() % 128;
	buff.y = rand() % 128;
	
	GFX_draw_line(buff.x, buff.y, buff.x + 2, buff.y + 2, ST7735_YELLOW);
}

void new_food(){
	normal_buff.x = rand() % 128;
	normal_buff.y = rand() % 128;
	
	LCD_draw_pixel(normal_buff.x, normal_buff.y, ST7735_GREEN);
}


void draw_obstacles(){
	obstacles[0].x1 = 10;
	obstacles[0].y1 = 10;
	obstacles[0].x2 = 90;
	obstacles[0].y2 = 10;
	
	obstacles[1].x1 = 110;
	obstacles[1].y1 = 40;
	obstacles[1].x2 = 110;
	obstacles[1].y2 = 100;
	
	obstacles[2].x1 = 40;
	obstacles[2].y1 = 45;
	obstacles[2].x2 = 40;
	obstacles[2].y2 = 75;
	
	obstacles[3].x1 = 40;
	obstacles[3].y1 = 45;
	obstacles[3].x2 = 20;
	obstacles[3].y2 = 45;

	obstacles[4].x1 = 20;
	obstacles[4].y1 = 45;
	obstacles[4].x2 = 20;
	obstacles[4].y2 = 75;
	
	int i;
	for(i = 0; i < NR_OBST; i++){
		GFX_draw_line(obstacles[i].x1, obstacles[i].y1,
						obstacles[i].x2, obstacles[i].y2, ST7735_BLACK);
	}
}

void draw_snake(){
	int i;
	
	for (i = 0; i < snake_length; i++) {
		LCD_draw_pixel(snake[i].x, snake[i].y, ST7735_RED);
	}
}

void snake_init(){
	int firstPosX = HEIGHT/2;
	int firstPosY = WIDTH/2;
	int i;
	
	for (i = 0; i < snake_length; i++) {
		snake[i].x = firstPosX + i;
		snake[i].y = firstPosY;
	}
	
	draw_snake();
	
}

void snake_movement(){
	int i;
	
	for (i = 0; i < snake_length - 1; i++) {
		LCD_draw_pixel(snake[i].x, snake[i].y, ST7735_BLUE);
		snake[i].x = snake[i + 1].x;
		snake[i].y = snake[i + 1].y;
	}
	
	if (snake_direction == UP) {
		if (snake[snake_length - 1].x + 1 == HEIGHT + 1)
			snake[snake_length - 1].x = 0;
		
		snake[snake_length - 1].x += 1;
	} else if (snake_direction == DOWN) {
		if (snake[snake_length - 1].x - 1 == 0)
			snake[snake_length - 1].x = HEIGHT + 1;
		
		snake[snake_length - 1].x -= 1;
	} else if (snake_direction == LEFT) {
		if (snake[snake_length - 1].y - 1 == 0)
			snake[snake_length - 1].y = WIDTH + 1;
		
		snake[snake_length - 1].y -= 1;
	} else if (snake_direction == RIGHT) {
		if (snake[snake_length - 1].y + 1 == WIDTH + 1)
			snake[snake_length - 1].y = 0;
		
		snake[snake_length - 1].y += 1;
	}
}

void snake_modify(int n){
	int i;
	int size = n;
	snake_length += size;
	
	
	if (snake_direction == UP) {
		for (i = size; i > 0; i--) {
			snake[snake_length - i].x = snake[snake_length - (i + 1)].x + 1;
			snake[snake_length - i].y = snake[snake_length - (i + 1)].y;
		}
	} else if (snake_direction == DOWN) {
		for (i = size; i > 0; i--) {
			snake[snake_length - i].x = snake[snake_length - (i + 1)].x - 1;
			snake[snake_length - i].y = snake[snake_length - (i + 1)].y;
		}
	} else if (snake_direction == LEFT) {
		for (i = size; i > 0; i--) {
			snake[snake_length - i].x = snake[snake_length - (i + 1)].x;
			snake[snake_length - i].y = snake[snake_length - (i + 1)].y - 1;
		}
	} else if (snake_direction == RIGHT) {
	for (i = size; i > 0; i--) {
			snake[snake_length - i].x = snake[snake_length - (i + 1)].x;
			snake[snake_length - i].y = snake[snake_length - (i + 1)].y + 1;
		}
	}
}

void game_over(){
	while (1) {
		LCD_fill_screen(ST7735_BLACK);
	}
}

void collision_detection(){
	int i;
	
	for (i = 0; i < NR_OBST; i++) {
		if (snake[snake_length - 1].x >= obstacles[i].x1 &&
			snake[snake_length - 1].x <= obstacles[i].x2)
			if (snake[snake_length - 1].y >= obstacles[i].y1 &&
				snake[snake_length - 1].y <= obstacles[i].y2)
				game_over();
	}
	
	/*for (i = 0; i < snake_length - 1; i++)
		if (snake[snake_length - 1].x == snake[i].x && snake[snake_length - 1].y == snake[i].y && (i != snake_length - 1))
			game_over();
	*/
	
	if (snake[snake_length - 1].x >= buff.x &&
		snake[snake_length - 1].x <= buff.x + 2)
		if (snake[snake_length - 1].y >= buff.y &&
			snake[snake_length - 1].y <= buff.y + 2) {
				GFX_draw_line(buff.x, buff.y, buff.x + 2, buff.y + 2, ST7735_BLUE);
				snake_modify(8);
				new_power_up();
		}
			
	
	if ((snake[snake_length - 1].x == normal_buff.x ||  snake[snake_length - 1].x == normal_buff.x + 1 ||  snake[snake_length - 1].x == normal_buff.x - 1) &&
			(snake[snake_length - 1].y == normal_buff.y || snake[snake_length - 1].y == normal_buff.y + 1 || snake[snake_length - 1].y == normal_buff.y - 1)) {
			LCD_draw_pixel(normal_buff.x, normal_buff.y, ST7735_BLUE);
			snake_modify(2);
			new_food();
	}
				
}

int main(void){
	LCD_init();	
	LCD_fill_screen(ST7735_BLUE);
	
	buttons_init();
	new_power_up();
	new_food();
	draw_obstacles();
	snake_init();

    while(1){
		if ((PIND & (1 << PD0)) == 0)
			snake_direction = RIGHT;
		else if ((PIND & (1 << PD1)) == 0)
			snake_direction = DOWN;
		else if ((PIND & (1 << PD4)) == 0)
			snake_direction = LEFT;
		else if ((PIND & (1 << PD5)) == 0)
			snake_direction = UP;
			
		snake_movement();
		draw_snake();
		collision_detection();
		if (snake_length > 50 && snake_length < 80)
			level = 700;
		else if(snake_length > 80 && snake_length < WIDTH)
			level = 300;
		_delay_ms(level);
    }

    return 0;
}
